#include <rte_config.h>
#include <rte_acl.h>
#include <rte_ip.h> //for IPv4 macro
#include <endian.h>
#include <assert.h>

#undef ALIGN
//#define ALIGN(x,a) (((x)+(a)-1)&~((a)-1))
#define ALIGN(x,a) ((x)%(a)==0 ? (x) : (x)+(a)-(x)%(a))

#define ACL_ALLOW 0

struct ipv4_5tuple {
    uint8_t fake_u8;
    uint8_t spare[3];
    uint32_t ip_src;
    uint32_t ip_dst;
    uint16_t port_src;
    uint16_t port_dst;
    uint8_t proto;
} __attribute__((packed));

enum {
	FAKE_U8_FIELD,
	SRC_FIELD_IPV4,
	DST_FIELD_IPV4,
	SRCP_FIELD_IPV4,
	DSTP_FIELD_IPV4,
	PROTO_FIELD_IPV4,
	NUM_FIELDS_IPV4
};

struct rte_acl_field_def ipv4_defs[NUM_FIELDS_IPV4] = {
	{
		.type = RTE_ACL_FIELD_TYPE_MASK,
		.size = sizeof(uint8_t),
		.field_index = FAKE_U8_FIELD,
		.input_index = 0,
		.offset = offsetof(struct ipv4_5tuple, fake_u8),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_MASK,
		.size = sizeof(uint32_t),
		.field_index = SRC_FIELD_IPV4,
		.input_index = 1,
		.offset = offsetof(struct ipv4_5tuple, ip_src),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_MASK,
		.size = sizeof(uint32_t),
		.field_index = DST_FIELD_IPV4,
		.input_index = 2,
		.offset = offsetof(struct ipv4_5tuple, ip_dst),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_RANGE,
		.size = sizeof(uint16_t),
		.field_index = SRCP_FIELD_IPV4,
		.input_index = 3,
		.offset = offsetof(struct ipv4_5tuple, port_src),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_RANGE,
		.size = sizeof(uint16_t),
		.field_index = DSTP_FIELD_IPV4,
		.input_index = 4,
		.offset = offsetof(struct ipv4_5tuple, port_dst),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_BITMASK,
		.size = sizeof(uint8_t),
		.field_index = PROTO_FIELD_IPV4,
		.input_index = 5,
		.offset = offsetof(struct ipv4_5tuple, proto),
	},
};

RTE_ACL_RULE_DEF(acl4_rule, RTE_DIM(ipv4_defs));

// TODO: Research if DPDK ACL support 64bit range.
#define FIELD_32prefix(point, prefix_len) {.value = {.u32 = point}, .mask_range = {.u32 = prefix_len}}
#define FIELD_32prefix_WILDCARD FIELD_32prefix(0, 0)

#define FIELD_32range(low, high) {.value = {.u32 = low}, .mask_range = {.u32 = high}}
#define FIELD_32range_WILDCARD FIELD_32range(0, (uint32_t)(~0U))

#define FIELD_16range(low, high) {.value = {.u16 = low}, .mask_range = {.u16 = high}}
#define FIELD_16range_WILDCARD FIELD_16range(0, (uint16_t)(~0U))

#define FIELD_8range(low, high) {.value = {.u8 = low}, .mask_range = {.u8 = high}}
#define FIELD_8range_WILDCARD FIELD_8range(0, 0xff)

#define FIELD_8bitmask(point, mask) {.value = {.u8 = point}, .mask_range = {.u8 = mask}}
#define FIELD_8bitmask_WILDCARD FIELD_8bitmask(0, 0x0)

#define FIELD_8prefix(point, mask) {.value = {.u8 = point}, .mask_range = {.u8 = mask}}
#define FIELD_8prefix_WILDCARD FIELD_8prefix(0, 0x0)

struct acl4_rule acl_test_rules[] = {
     {
          .data = {.userdata = 1, .category_mask = 0x1,
                   .priority = 2},
          .field = {
	       FIELD_8prefix_WILDCARD,
               FIELD_32prefix_WILDCARD,
               FIELD_32prefix_WILDCARD,
               FIELD_16range_WILDCARD,
               FIELD_16range_WILDCARD,
	       FIELD_8bitmask(1, 0x1),
          }
     },
};

struct ipv4_5tuple acl_test_data[] = {
     {.proto=17, .ip_src = IPv4(192,168,0,1), .ip_dst = IPv4(192,168,0,1), .port_src = 49, .port_dst = 1001}, /* should match 1 */
     {.proto=17, .ip_src = IPv4(192,168,0,4), .ip_dst = IPv4(192,168,0,4), .port_src = 49, .port_dst = 1001}, /* should match 1 */
};

/* AC context creation parameters. */
struct rte_acl_param prm = {
    .name = "ACL_8e",
    .socket_id = SOCKET_ID_ANY,
    .rule_size = RTE_ACL_RULE_SZ(RTE_DIM(ipv4_defs)),

    /* number of fields per rule. */
    .max_rule_num = 4096, /* maximum number of rules in the AC context. */
};

struct rte_acl_config cfg = {
     .num_categories = 1,
     .num_fields = RTE_DIM(ipv4_defs),
};


struct rte_acl_ctx * acx = NULL;

static int app_acl_init(void) {
     int rc = 0;
/* create an empty AC context  */
     if ((acx = rte_acl_create(&prm)) == NULL) {
          /* handle context create failure. */
          return -1;
     }
     
     int i = 0;
     for(i=0; i<RTE_DIM(acl_test_rules); i++){
/* add rules to the context */
          rc = rte_acl_add_rules(acx, (struct rte_acl_rule *)&acl_test_rules[i], 1);
          if (rc != 0) {
               /* handle error at adding ACL rules. */
               return -1;
          }
     }

/* prepare AC build config. */
     memcpy(cfg.defs, ipv4_defs, sizeof (ipv4_defs));
     
/* build the runtime structures for added rules, with 2 categories. */
     rc = rte_acl_build(acx, &cfg);
     rte_acl_dump(acx);

     if (rc != 0) {
          /* handle error at build runtime structures for ACL context. */
          return -1;
     }
     return 0;
}

int main(int argc, char **argv){
     int rc;
     int dpdk_argc = 5;
     char *dpdk_argv[] = {argv[0], "-c", "1", "-n", "4"};

     rc = rte_eal_init(dpdk_argc, dpdk_argv);
     if (rc < 0)
          rte_panic("Cannot init EAL\n");

     int i = 0;
     rc = app_acl_init();
     if(rc!=0){
          printf("app_acl_init failed!\n");
          return -1;
     }

/* Note When adding new rules into an ACL context, all fields must be in host byte order (LSB). When the search is performed for an input tuple, all fields in that tuple must be in network byte order (MSB). */
     for(i=0; i<RTE_DIM(acl_test_data); i++){
          acl_test_data[i].ip_src = htobe32(acl_test_data[i].ip_src);
          acl_test_data[i].ip_dst = htobe32(acl_test_data[i].ip_dst);
          acl_test_data[i].port_src = htobe16(acl_test_data[i].port_src);
          acl_test_data[i].port_dst = htobe16(acl_test_data[i].port_dst);
     }

     uint8_t **data = (uint8_t **)calloc(RTE_DIM(acl_test_data), sizeof(uint8_t *));
     for(i=0; i<RTE_DIM(acl_test_data); i++){
          data[i] = (uint8_t *)&acl_test_data[i];
          int keydata_len = sizeof(struct ipv4_5tuple);
          printf("keyData(%d bytes): ", keydata_len);
          int j = 0;
          for(j=0; j<keydata_len; j++){
               printf("%02x", data[i][j]);
          }
          printf("\n");
     }

/* categories parameter should be either one or multiple of RTE_ACL_RESULTS_MULTIPLIER and can't be bigger than RTE_ACL_MAX_CATEGORIES. */
     /* round up to nearest RTE_ACL_RESULTS_MULTIPLIER */
     uint32_t categories = cfg.num_categories==1 ? 1 : ALIGN(cfg.num_categories, RTE_ACL_RESULTS_MULTIPLIER);
     printf("cfg.num_categories=%d, categories=%d\n", cfg.num_categories, categories);
     uint32_t *results = (uint32_t *)calloc(RTE_DIM(acl_test_data), categories*sizeof(uint32_t));
     rc = rte_acl_classify(acx, (const uint8_t **)data, results, RTE_DIM(acl_test_data), categories);
     if(rc!=0)
          printf("rc=%d, EINVAL=%d\n", rc, EINVAL);
     assert(rc==0);

     /* verify lookup result matchs the expectation  */
     for(i=0; i<RTE_DIM(acl_test_data); i++){
          printf("data %d, result:(%d), expect:(%d)\n", i, results[i*categories + ACL_ALLOW], 1);
          assert(results[i*categories + ACL_ALLOW] == 1);
     }
     
     free(results);
     /* free context */
     rte_acl_free(acx);

     printf("DPDK ACL test passed.\n");
     return 0;
}
