#include "acl_mp.h"
#include <endian.h>
#include <assert.h>

struct rte_acl_ctx * acx = NULL;

static int app_acl_init(void) {
     int rc = 0;
/* create an empty AC context  */
     if ((acx = rte_acl_create(&prm)) == NULL) {
          /* handle context create failure. */
          return -1;
     }
     
     int i = 0;
     for(i=0; i<RTE_DIM(acl_test_rules); i++){
/* add rules to the context */
          rc = rte_acl_add_rules(acx, (struct rte_acl_rule *)&acl_test_rules[i], 1);
          if (rc != 0) {
               /* handle error at adding ACL rules. */
               return -1;
          }
     }

/* prepare AC build config. */
     memcpy(cfg.defs, ipv4_defs, sizeof (ipv4_defs));
     
/* build the runtime structures for added rules, with 2 categories. */
     rc = rte_acl_build(acx, &cfg);
     rte_acl_dump(acx);

     if (rc != 0) {
          /* handle error at build runtime structures for ACL context. */
          return -1;
     }
     return 0;
}

int main(int argc, char **argv){
     int rc;
     int dpdk_argc = 5;
     char *dpdk_argv[] = {argv[0], "-c", "1", "-n", "4"};

     rc = rte_eal_init(dpdk_argc, dpdk_argv);
     if (rc < 0)
          rte_panic("Cannot init EAL\n");

     int i = 0;
     rc = app_acl_init();
     if(rc!=0){
          printf("app_acl_init failed!\n");
          return -1;
     }

/* Note When adding new rules into an ACL context, all fields must be in host byte order (LSB). When the search is performed for an input tuple, all fields in that tuple must be in network byte order (MSB). */
     for(i=0; i<RTE_DIM(acl_test_data); i++){
          acl_test_data[i].ip_src = htobe32(acl_test_data[i].ip_src);
          acl_test_data[i].ip_dst = htobe32(acl_test_data[i].ip_dst);
          acl_test_data[i].port_src = htobe16(acl_test_data[i].port_src);
          acl_test_data[i].port_dst = htobe16(acl_test_data[i].port_dst);
     }

     uint8_t **data = (uint8_t **)calloc(RTE_DIM(acl_test_data), sizeof(uint8_t *));
     for(i=0; i<RTE_DIM(acl_test_data); i++){
          data[i] = (uint8_t *)&acl_test_data[i];
     }

/* categories parameter should be either one or multiple of RTE_ACL_RESULTS_MULTIPLIER and can't be bigger than RTE_ACL_MAX_CATEGORIES. */
     /* round up to nearest RTE_ACL_RESULTS_MULTIPLIER */
     uint32_t categories = cfg.num_categories==1 ? 1 : ALIGN(cfg.num_categories, RTE_ACL_RESULTS_MULTIPLIER);
     printf("cfg.num_categories=%d, categories=%d\n", cfg.num_categories, categories);
     uint32_t *results = (uint32_t *)calloc(RTE_DIM(acl_test_data), categories*sizeof(uint32_t));
     rc = rte_acl_classify(acx, (const uint8_t **)data, results, RTE_DIM(acl_test_data), categories);
     if(rc!=0)
          printf("rc=%d, EINVAL=%d\n", rc, EINVAL);
     assert(rc==0);

/*     uint32_t *results = (uint32_t *)calloc(1, cfg.num_categories*sizeof(uint32_t));
     rc = rte_acl_classify(acx, (const uint8_t **)data, results, 1, cfg.num_categories);
     if(rc!=0)
          printf("rc=%d, EINVAL=%d, RTE_ACL_RESULTS_MULTIPLIER=%d\n", rc, EINVAL, RTE_ACL_RESULTS_MULTIPLIER);
     assert(rc==0);
*/
     
     // sleep until user type enter so that acl_mp2 is able to lookup.
     int c = getchar();

     /* verify lookup result matchs the expectation  */
     for(i=0; i<RTE_DIM(acl_test_data); i++){
          printf("data %d, result:(%d, %d), expect:(%d, %d)\n", i, results[i*categories + ACL_ALLOW], results[i*categories + ACL_DENY], acl_test_data[i].allow, acl_test_data[i].deny);
          assert(results[i*categories + ACL_ALLOW] == acl_test_data[i].allow);
          assert(results[i*categories + ACL_DENY] == acl_test_data[i].deny);
     }
     
     free(results);
     /* free context */
     rte_acl_free(acx);

     printf("DPDK ACL test passed.\n");
     return 0;
}
