#ifndef ACL_MP
#define ACL_MP
#include <rte_config.h>
#include <rte_acl.h>
#include <rte_ip.h> //for IPv4 macro

/* refers to dpdk-1.7.0/examples/l3fwd-acl/main.c, dpdk-1.7.0/app/test-acl/main.c, dpdk-1.7.0/app/test/test_acl.h. */

#undef ALIGN
//#define ALIGN(x,a) (((x)+(a)-1)&~((a)-1))
#define ALIGN(x,a) ((x)%(a)==0 ? (x) : (x)+(a)-(x)%(a))

#define ACL_ALLOW 0
#define ACL_DENY 1
#define ACL_ALLOW_MASK 0x1
#define ACL_DENY_MASK  0x2

struct ipv4_5tuple {
    uint8_t proto;
    uint16_t spare;
    uint32_t ip_src;
    uint32_t ip_dst;
    uint16_t port_src;
    uint16_t port_dst;
    uint32_t allow; //expected lookup result of ALLOW category
    uint32_t deny;  //expected lookup result of DENY category
} __attribute__((packed));

enum {
	PROTO_FIELD_IPV4,
	SRC_FIELD_IPV4,
	DST_FIELD_IPV4,
	SRCP_FIELD_IPV4,
	DSTP_FIELD_IPV4,
	NUM_FIELDS_IPV4
};


struct rte_acl_field_def ipv4_defs[NUM_FIELDS_IPV4] = {
	{
		.type = RTE_ACL_FIELD_TYPE_BITMASK,
		.size = sizeof(uint8_t),
		.field_index = PROTO_FIELD_IPV4,
		.input_index = 0,
		.offset = 0,
	},
	{
		.type = RTE_ACL_FIELD_TYPE_MASK,
		.size = sizeof(uint32_t),
		.field_index = SRC_FIELD_IPV4,
		.input_index = 1,
		.offset = offsetof(struct ipv4_5tuple, ip_src),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_MASK,
		.size = sizeof(uint32_t),
		.field_index = DST_FIELD_IPV4,
		.input_index = 2,
		.offset = offsetof(struct ipv4_5tuple, ip_dst),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_RANGE,
		.size = sizeof(uint16_t),
		.field_index = SRCP_FIELD_IPV4,
		.input_index = 3,
		.offset = offsetof(struct ipv4_5tuple, port_src),
	},
	{
		.type = RTE_ACL_FIELD_TYPE_RANGE,
		.size = sizeof(uint16_t),
		.field_index = DSTP_FIELD_IPV4,
		.input_index = 4,
		.offset = offsetof(struct ipv4_5tuple, port_dst),
	},
};

RTE_ACL_RULE_DEF(acl4_rule, RTE_DIM(ipv4_defs));

#define FIELD_32p(point, prefix_len) {.value = {.u32 = point}, .mask_range = {.u32 = prefix_len}}
#define FIELD_32p_WILDCARD FIELD_32p(0, 0)

// TODO: Research if DPDK ACL support 64bit range.
// Note: OPPL use 8r, 16r, 32r, 128p.
#define FIELD_64r(low, high) {.value = {.u64 = low}, .mask_range = {.u64 = high}}
#define FIELD_64r_WILDCARD FIELD_64r(0, (uint64_t)(~0ULL))

#define FIELD_32r(low, high) {.value = {.u32 = low}, .mask_range = {.u32 = high}}
#define FIELD_32r_WILDCARD FIELD_32r(0, (uint32_t)(~0U))

#define FIELD_16r(low, high) {.value = {.u16 = low}, .mask_range = {.u16 = high}}
#define FIELD_16r_WILDCARD FIELD_16r(0, (uint16_t)(~0U))

#define FIELD_8r(low, high) {.value = {.u8 = low}, .mask_range = {.u8 = high}}
#define FIELD_8r_WILDCARD FIELD_8r(0, 0xff)

#define FIELD_8m(point, mask) {.value = {.u8 = point}, .mask_range = {.u8 = mask}}
#define FIELD_8m_WILDCARD FIELD_8m(0, 0x0)

struct acl4_rule acl_test_rules[] = {
/* destination IP addresses */
     /* matches all packets traveling to 192.168.0.0/16 */
     {
          .data = {.userdata = 1, .category_mask = ACL_ALLOW_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p(IPv4(192, 168, 0, 0), 16),
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD
          }
     },
     /* matches all packets traveling to 192.168.1.0/24 */
     {
          .data = {.userdata = 2, .category_mask = ACL_ALLOW_MASK,
                   .priority = 3},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p(IPv4(192, 168, 1, 0), 24),
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD
          }
     },
     /* matches all packets traveling to 192.168.1.50 */
     {
          .data = {.userdata = 3, .category_mask = ACL_DENY_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p(IPv4(192, 168, 1, 50), 32),
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD
          }
     },
/* source IP addresses */
     /* matches all packets traveling from 10.0.0.0/8 */
     {
          .data = {.userdata = 4, .category_mask = ACL_ALLOW_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p(IPv4(10, 0, 0, 0), 8),
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD
          }
     },
     /* matches all packets traveling from 10.1.1.0/24 */
     {
          .data = {.userdata = 5, .category_mask = ACL_ALLOW_MASK,
                   .priority = 3},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p(IPv4(10, 1, 1, 0), 24),
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD
          }
     },
     /* matches all packets traveling from 10.1.1.1 */
     {
          .data = {.userdata = 6, .category_mask = ACL_DENY_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p(IPv4(10, 1, 1, 1), 32),
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD
          }
     },
/* destination port */
     /* matches everything with dst port 80 */
     {
          .data = {.userdata = 13, .category_mask = ACL_ALLOW_MASK,
                   .priority = 3},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r(80, 80)
          }
     },
     /* matches everything with dst port 22-1023 */
     {
          .data = {.userdata = 14, .category_mask = ACL_ALLOW_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r(22, 1023)
          }
     },
     /* matches everything with dst port 1020 */
     {
          .data = {.userdata = 15, .category_mask = ACL_DENY_MASK,
                   .priority = 3},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r(1020, 1020)
          }
     },
     /* matches everything with dst portrange  1000-2000 */
     {
          .data = {.userdata = 16, .category_mask = ACL_DENY_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r(1000, 2000)
          }
     },
     /* matches everything with src port 80 */
     {
          .data = {.userdata = 17, .category_mask = ACL_ALLOW_MASK,
                   .priority = 3},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r(80, 80),
               FIELD_16r_WILDCARD,
          }
     },
     /* matches everything with src port 22-1023 */
     {
          .data = {.userdata = 18, .category_mask = ACL_ALLOW_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r(22, 1023),
               FIELD_16r_WILDCARD,
          }
     },
     /* matches everything with src port 1020 */
     {
          .data = {.userdata = 19, .category_mask = ACL_DENY_MASK,
                   .priority = 3},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r(1020, 1020),
               FIELD_16r_WILDCARD,
          }
     },
     /* matches everything with src portrange  1000-2000 */
     {
          .data = {.userdata = 20, .category_mask = ACL_DENY_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r(1000, 2000),
               FIELD_16r_WILDCARD,
          }
     },
/* protocol number */
     /* matches all packets with protocol number either 0x64 or 0xE4 */
     {
          .data = {.userdata = 21, .category_mask = ACL_ALLOW_MASK,
                   .priority = 2},
          .field = {
               FIELD_8m(0x64, 0x7f),
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD,
          }
     },
     /* matches all packets with protocol that have 0x5 in them */
     {
          .data = {.userdata = 22, .category_mask = ACL_ALLOW_MASK,
                   .priority = 1},
          .field = {
               FIELD_8m(0x5, 0x5),
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD,
          }
     },
     /* matches all packets with protocol 5 */
     {
          .data = {.userdata = 23, .category_mask = ACL_DENY_MASK,
                   .priority = 3},
          .field = {
               FIELD_8m(0x5, 0xff),
               FIELD_32p_WILDCARD,
               FIELD_32p_WILDCARD,
               FIELD_16r_WILDCARD,
               FIELD_16r_WILDCARD,
          }
     },
/* rules combining various fields */
     {
          .data = {.userdata = 24, .category_mask = ACL_ALLOW_MASK,
                   .priority = 4},
          .field = {
               FIELD_8m(0x5, 0xff),
               FIELD_32p(IPv4(5,6,7,8), 24),
               FIELD_32p(IPv4(1,2,3,4), 16),
               FIELD_16r_WILDCARD,
               FIELD_16r(22, 1024)
          }
     },
     {
          .data = {.userdata = 25, .category_mask = ACL_DENY_MASK,
                   .priority = 4},
          .field = {
               FIELD_8m(0x5, 0xff),
               FIELD_32p(IPv4(1,2,3,4), 16),
               FIELD_32p(IPv4(5,6,7,8), 24),
               FIELD_16r_WILDCARD,
               FIELD_16r(22, 1024)
          }
     },
     {
          .data = {.userdata = 26, .category_mask = ACL_ALLOW_MASK,
                   .priority = 5},
          .field = {
               FIELD_8m(0x5, 0xff),
               FIELD_32p(IPv4(5,6,7,8), 32),
               FIELD_32p(IPv4(1,2,3,4), 8),
               FIELD_16r_WILDCARD,
               FIELD_16r(22, 1024)
          }
     },
     {
          .data = {.userdata = 27, .category_mask = ACL_DENY_MASK,
                   .priority = 5},
          .field = {
               FIELD_8m(0x5, 0xff),
               FIELD_32p(IPv4(1,2,3,4), 8),
               FIELD_32p(IPv4(5,6,7,8), 32),
               FIELD_16r_WILDCARD,
               FIELD_16r(22, 1024)
          }
     },
};

struct ipv4_5tuple acl_test_data[] = {
/* testing single rule aspects */
		{.ip_src = IPv4(10,0,0,0), .allow = 4}, /* should match 4 */
		{.ip_src = IPv4(10,1,1,2), .allow = 5}, /* should match 5 */
		{.ip_src = IPv4(10,1,1,1), .allow = 5,
				.deny = 6},                     /* should match 5, 6 */
		{.ip_dst = IPv4(10,0,0,0)},             /* should not match */
		{.ip_dst = IPv4(10,1,1,2)},             /* should not match */
		{.ip_dst = IPv4(10,1,1,1)},             /* should not match */

		{.ip_src = IPv4(192,168,2,50)},             /* should not match */
		{.ip_src = IPv4(192,168,1,2)},              /* should not match */
		{.ip_src = IPv4(192,168,1,50)},             /* should not match */
		{.ip_dst = IPv4(192,168,2,50), .allow = 1}, /* should match 1 */
		{.ip_dst = IPv4(192,168,1,49), .allow = 2}, /* should match 2 */
		{.ip_dst = IPv4(192,168,1,50), .allow = 2,
				.deny = 3},                         /* should match 2, 3 */

		{.port_dst = 80, .allow = 13},                /* should match 13 */
		{.port_dst = 79, .allow = 14},                /* should match 14 */
		{.port_dst = 81, .allow = 14},                /* should match 14 */
		{.port_dst = 21},                             /* should not match */
		{.port_dst = 1024, .deny = 16},               /* should match 16 */
		{.port_dst = 1020, .allow = 14, .deny = 15},  /* should match 14, 15 */

		{.port_src = 80, .allow = 17},                /* should match 17 */
		{.port_src = 79, .allow = 18},                /* should match 18 */
		{.port_src = 81, .allow = 18},                /* should match 18 */
		{.port_src = 21},                             /* should not match */
		{.port_src = 1024, .deny = 20},               /* should match 20 */
		{.port_src = 1020, .allow = 18, .deny = 19},  /* should match 18, 19 */

		{.proto = 0x64, .allow = 21},             /* should match 21 */
		{.proto = 0xE4, .allow = 21},             /* should match 21 */
		{.proto = 0xE2},                          /* should not match */
		{.proto = 0xD, .allow = 22},              /* should match 22 */
		{.proto = 0x6},                           /* should not match */
		{.proto = 0x5, .allow = 22, .deny = 23},  /* should match 22, 23 */

/* testing matching multiple rules at once */
		{.port_dst = 80, .port_src = 1024,
				.allow = 13, .deny = 20},             /* should match 13,20 */
		{.port_dst = 79, .port_src = 1024,
				.allow = 14, .deny = 20},             /* should match 14,20 */
		{.proto = 0x5, .ip_dst = IPv4(192,168,2,50),
				.allow = 1, .deny = 23},               /* should match 1, 23 */

		{.proto = 0x5, .ip_dst = IPv4(192,168,1,50),
				.allow = 2, .deny = 23},              /* should match 2, 23 */
		{.proto = 0x5, .port_src = 80,
				.allow = 17, .deny = 23},             /* should match 17, 23 */
		{.proto = 0x5, .port_dst = 80,
				.allow = 13, .deny = 23},             /* should match 13, 23 */
		{.proto = 0x51, .port_src = 5000},            /* should not match */
		{.ip_src = IPv4(192,168,1,50),
				.ip_dst = IPv4(10,0,0,0),
				.proto = 0x51,
				.port_src = 5000,
				.port_dst = 5000},                    /* should not match */

/* test full packet rules */
		{
				.ip_dst = IPv4(1,2,100,200),
				.ip_src = IPv4(5,6,7,254),
				.proto = 0x5,
				.port_src = 12345,
				.port_dst = 80,
				.allow = 24,
				.deny = 23
		}, /* should match 23, 24 */
		{
				.ip_dst = IPv4(5,6,7,254),
				.ip_src = IPv4(1,2,100,200),
				.proto = 0x5,
				.port_src = 12345,
				.port_dst = 80,
				.allow = 13,
				.deny = 25
		}, /* should match 13, 25 */
		{
				.ip_dst = IPv4(1,10,20,30),
				.ip_src = IPv4(5,6,7,8),
				.proto = 0x5,
				.port_src = 12345,
				.port_dst = 80,
				.allow = 26,
				.deny = 23
		}, /* should match 26, 23 */
		{
				.ip_dst = IPv4(5,6,7,8),
				.ip_src = IPv4(1,10,20,30),
				.proto = 0x5,
				.port_src = 12345,
				.port_dst = 80,
				.allow = 13,
				.deny = 27
		}, /* should match 13, 27 */
		{
				.ip_dst = IPv4(2,2,3,4),
				.ip_src = IPv4(4,6,7,8),
				.proto = 0x5,
				.port_src = 12345,
				.port_dst = 80,
				.allow = 13,
				.deny = 23
		}, /* should match 13, 23 */
		{
				.ip_dst = IPv4(1,2,3,4),
				.ip_src = IPv4(4,6,7,8),
				.proto = 0x5,
				.port_src = 12345,
				.port_dst = 80,
				.allow = 13,
				.deny = 23
		}, /* should match 13, 23 */

		{
				.ip_dst = IPv4(1,2,3,4),
				.ip_src = IPv4(4,6,7,8),
				.proto = 0x50,
				.port_src = 12345,
				.port_dst = 12345,
		}, /* should not match */
};

/* AC context creation parameters. */
struct rte_acl_param prm = {
    .name = "ACL_example",
    .socket_id = SOCKET_ID_ANY,
    .rule_size = RTE_ACL_RULE_SZ(RTE_DIM(ipv4_defs)),

    /* number of fields per rule. */
    .max_rule_num = 4096, /* maximum number of rules in the AC context. */
};

struct rte_acl_config cfg = {
     .num_categories = 2,
     .num_fields = RTE_DIM(ipv4_defs),
};

          

#endif
